from .controllers import PluginController, RgbLampController, DeviceController
from .controllers import Controller
from .controllers import TvController
from .controllers import AcController
from .controllers import MultiCommandDeviceController
from .controllers import MultiCommandRgbLampController
from .models import PluginsModel, RgbLampsModel, RgbLampStatesModel, DevicesModel
from .models import TvsModel, TvStatesModel
from .models import AcsModel, AcStatesModel
from .models import MultiCommandDevicesModel, CommandsModel
from .models import MultiCommandRgbLampsModel
from .models import DeviceStatesModel, clone
from .managers import Manager, PluginManager



from .device_model import DevicesModel, DeviceStatesModel
from durc import clone


class AcStatesModel(DeviceStatesModel):
    schema = clone(DeviceStatesModel.schema) | {
        'temp': { 'type': 'integer', 'default': 16 },
    }
    collection = DeviceStatesModel.collection


class AcsModel(DevicesModel):
    schema = clone(DevicesModel.schema) | {
        'temp_max': { 'type': 'integer', 'default': 24 },
        'temp_min': { 'type': 'integer', 'default': 16 },
    }
    children_model_classes = clone(DevicesModel.children_model_classes)
    children_model_classes['state']['class'] = AcStatesModel



from .rgb_light_model import RgbLampsModel
from .multi_command_devices_model import MultiCommandDevicesModel
from durc import Model, clone


class ColorCommandsModel(Model):
    schema = clone(Model.schema) | {
        'color_id': { 'type': 'string', 'required': True },
        'command_id': { 'type': 'string', 'required': True },
    }


class MultiCommandRgbLampsModel(MultiCommandDevicesModel, RgbLampsModel):
    schema = clone(RgbLampsModel.schema) | {
        'brightness_max': { 'type': 'integer', 'default': 100 },
        'brightness_min': { 'type': 'integer', 'default': 0 },
    }
    children_model_classes = clone(MultiCommandDevicesModel.children_model_classes)
    children_model_classes |= clone(RgbLampsModel.children_model_classes) | {
        'color_commands': { 'class': ColorCommandsModel, 'quantity': '*' },
    }



from .device_model import DevicesModel, DeviceStatesModel
from durc import clone


class RgbLampStatesModel(DeviceStatesModel):
    schema = clone(DeviceStatesModel.schema) | {
        'color': { 'type': 'string', 'default': '#ffffff' },
        'brightness': { 'type': 'number', 'default': 0 },
    }
    collection = DeviceStatesModel.collection


class RgbLampsModel(DevicesModel):
    schema = clone(DevicesModel.schema)
    children_model_classes = clone(DevicesModel.children_model_classes)
    children_model_classes['state']['class'] = RgbLampStatesModel



from abc import abstractmethod
from ..models import TvsModel
from .device_controller import DeviceController

class TvController(DeviceController):
    model_class = TvsModel


    @abstractmethod
    def mute(self):
        pass


    @abstractmethod
    def set_vol_by(self, n):
        pass


    @abstractmethod
    def set_vol_to(self, target):
        pass



from abc import abstractmethod
from ..models import AcsModel
from .device_controller import DeviceController

class AcController(DeviceController):
    model_class = AcsModel


    @abstractmethod
    def set_temp_by(self, n):
        pass


    @abstractmethod
    def set_temp_to(self, target):
        pass



from abc import abstractmethod
from ..models import MultiCommandDevicesModel
from .device_controller import DeviceController

class MultiCommandDeviceController(DeviceController):
    model_class = MultiCommandDevicesModel


    @abstractmethod
    def send_command(self, id):
        pass


    def add_command(self, id, data):
        pass


    def remove_command(self, id):
        pass


    def update_command(self, id, data):
        pass


    def set_int_state_attr_to(self, attr, f):
        self.set_int_state_attr_by(attr, int(f) - getattr(self.model.state, attr))


    def set_int_state_attr_by(self, attr, n=0, delta=0):
        curr_val = getattr(self.model.state, attr) + delta
        max_val = int(getattr(self.model, f'{attr}_max'))
        min_val = int(getattr(self.model, f'{attr}_min'))
        n = int(n)
        if (n > 0 and curr_val < max_val) or (n < 0 and curr_val > min_val):
            cmd = f'{attr}_up' if n > 0 else f'{attr}_down'
            self.send_command(cmd)
            inc = 1 if n > 0 else -1
            self.set_int_state_attr_by(attr, n - inc, delta + inc)
            return True
        if delta:
            self.set_state(**{attr: curr_val})
        return False



from abc import ABC, abstractmethod
from ..utils import get_base_classes, get_methods_in
from ..models import DevicesModel
from . import Controller
import os

class DeviceController(Controller, ABC):
    model_class = DevicesModel
    manager_class = None


    def __init__(self):
        super().__init__()
        self._dev = None


    @property
    def name(self): return self.model.name


    @property
    def state(self): return self.model.state


    @abstractmethod
    def on(self):
        # This function must return True | False
        pass


    @abstractmethod
    def off(self):
        # This function must return True | False
        pass


    def toggle(self):
        # This function must return True | False
        return self.off() if self.is_on() else self.on()


    def get_power(self):
        # This function must return on | off
        ping = self.model.power_by_ping
        if self.model.addr and ping:
            addr = self.model.addr
            return 'off' if os.system(f'fping -c1 -t100 {addr} &>/dev/null') else 'on'
        return self.model.state.power


    def get_switch_power(self):
        return 'on' if self.model.state.power == 'off' else 'off'


    @property
    def dev(self): return self._dev


    @dev.setter
    def dev(self, d): self._dev = d


    def on_load(self, **data):
        if self.manager_class:
            self.manager_class.add_device(self)
        self.sync_state()


    def sync_state(self):
        # will execute getters of highest hierarchy class if they exist.
        # Its important to notice that the controllers must implement the getters
        # with the 'get_' prefix
        actions = self.get_actions()
        state = {}
        for k in self.model.state.schema.keys():
            getter_id = f'get_{k}'
            if getter_id in actions:
                state[k] = actions[getter_id](self)
        self.set_state(**state)


    def set_state(self, **data):
        self.model.state.update(**data)


    def is_on(self):
        return self.get_power() == 'on'


    def should_update_power(self, action):
        is_on = self.is_on()
        return action == 'toggle' or \
            (action == 'off' and is_on) or \
            (action == 'on' and not is_on)


    def on_power_changed(self, new_state):
        self.set_state(power=new_state)


    def get_actions(self):
        bases = get_base_classes(self.__class__)
        methods = get_methods_in(*bases)
        actions = {}
        for k, f in methods:
            if k not in actions:
                actions[k] = f
        return actions


    def trigger_action(self, action_id, *params):
        try:
            actions = self.get_actions()
            if action_id in ['on', 'off', 'toggle']:
                if self.should_update_power(action_id):
                    new_state = self.get_switch_power()
                    if actions[action_id](self):
                        self.on_power_changed(new_state)
                    return True
                else:
                    return False
            if action_id in actions:
                actions[action_id](self, *params)
            return True
        except:
            return False



from durc import Model, clone
from .device_model import DevicesModel


class CommandsModel(Model):
    schema = {
        **Model.schema,
        'name': { 'type': 'string' },
        'label': { 'type': 'string', 'default': '' },
        'data': { 'type': 'string', 'default': '', 'required': True },
    }


class MultiCommandDevicesModel(DevicesModel):
    children_model_classes = clone(DevicesModel.children_model_classes) | {
        'commands': { 'class': CommandsModel, 'quantity': '*' }
    }
